import logging
import aiohttp
import aiohttp.web
import aiomysql.sa

# get logger
logger = logging.getLogger("pdnsexporter")

# init web router
router = aiohttp.web.RouteTableDef()

# middleware for basic auth
@aiohttp.web.middleware
async def checkauth(request, handler):
    # read authorization header
    auth = request.headers.get('Authorization')
    if auth is None: 
        return aiohttp.web.Response(status=401)

    # decode the basic authorization
    basicauth = aiohttp.BasicAuth(login="").decode(auth_header=auth)
    if basicauth.login != request.app["cfg"]["api-login"]:
        return aiohttp.web.Response(status=401)

    if basicauth.password != request.app["cfg"]["api-password"]:
        return aiohttp.web.Response(status=401)

    return await handler(request)

@router.get("/zones")
async def list_zones(request):
    """return all zones"""
    # get domains from database
    async with request.app['db'].acquire() as conn:
        cursor = await conn.execute("SELECT id,name FROM domains;")
        rows = await cursor.fetchall()
        zones = [dict(q) for q in rows]

    return aiohttp.web.json_response(zones)

@router.get("/zone/{zoneid}")
async def view_zone(request):
    """return content zone in bind format"""
    zone_id = request.match_info["zoneid"]

    # get dns records from database
    async with request.app['db'].acquire() as conn:
        cursor = await conn.execute("SELECT name, type, content, ttl FROM records WHERE domain_id=%s;", (zone_id,))
        rows = await cursor.fetchall()
        records = [dict(q) for q in rows]

    # prepare the zone file
    zone = [ "$ORIGIN ." ]
    for r in records:
        zone.append( "%s\t%s\tIN\t%s\t%s" % (r["name"], r["ttl"], r["type"], r["content"]) )
    zone.append("")

    return aiohttp.web.Response(text="\n".join(zone), content_type='text/plain')

@router.get("/metrics")
async def view_metrics(request):
    """return metrics for prometheus"""
    rtypes = ["soa", "a", "aaaa", "txt", "cname", "ptr", "mx", "ns", "srv", "ptr", "mx"]

    # fetch domains and records from database
    async with request.app['db'].acquire() as conn:
        cursor = await conn.execute("SELECT id,name FROM domains;")
        rows = await cursor.fetchall()
        zones = [dict(q) for q in rows]

    # iter on each zones to fetch records
    for zone in zones:
        async with request.app['db'].acquire() as conn:
            cursor = await conn.execute("SELECT name, type, content, ttl FROM records WHERE domain_id=%s;", (zone["id"],))
            rows = await cursor.fetchall()
            records = [dict(q) for q in rows]

            # add some stats in zone
            zone["records"] = len(records)
            for rtype in rtypes:
                zone["records_%s" % rtype] = len([r for r in records if r.get('type')==rtype.upper()])

    # prepare metrics
    metrics = []
    metrics.append( "# HELP pdnsexporter_zones_total Total number of zones" )
    metrics.append( "# TYPE pdnsexporter_zones_total counter" )
    metrics.append( "pdnsexporter_zones_total %s" % len(zones) )

    metrics.append( "# HELP pdnsexporter_records_total Total number of records" )
    metrics.append( "# TYPE pdnsexporter_records_total counter" )
    metrics.append( "pdnsexporter_records_total %s" % sum(z.get('records')  for z in zones) )

    for rtype in rtypes:
        metrics.append( "# HELP pdnsexporter_records_%s_total Total number of records %s" % (rtype, rtype.upper()) )
        metrics.append( "# TYPE pdnsexporter_records_%s_total counter" % rtype )
        metrics.append( "pdnsexporter_records_%s_total %s" % (rtype, sum(z.get('records_%s' % rtype)  for z in zones)) )

    for zone in zones:
        metrics.append( "# HELP pdnsexporter_records Number of records" )
        metrics.append( "# TYPE pdnsexporter_records counter" )
        for rtype in rtypes:
            metrics.append( "# HELP pdnsexporter_records_%s Number of records %s" % (rtype, rtype.upper()) )
            metrics.append( "# TYPE pdnsexporter_records_%s counter" % rtype )

        metrics.append( "pdnsexporter_records{identity=\"%s\"} %s" % (zone["name"], zone["records"]) )
        for rtype in rtypes:
            metrics.append( "pdnsexporter_records_%s{identity=\"%s\"} %s" % (rtype, zone["name"], zone["records_%s" % rtype]) )

    metrics.append( "" )

    return aiohttp.web.Response(text="\n".join(metrics), content_type='text/plain')

async def open_db(app):
    """create engine database"""
    engine = await aiomysql.sa.create_engine(db=app["cfg"]['db-pdns-name'],
                                             user=app["cfg"]['db-pdns-user'],
                                             password=app["cfg"]['db-pdns-password'],
                                             host=app["cfg"]['db-pdns-host'],
                                             port=int(app["cfg"]['db-pdns-port']))
    app["db"] = engine

async def close_db(app):
    """close properly all connections with database"""
    app['db'].close()
    await app['db'].wait_closed()

def setup_webapp(cfg):
    """setup the web application"""
    # init the web application
    app = aiohttp.web.Application(middlewares=[checkauth])

    # setup routes
    app.add_routes(router)

    # add config
    app["cfg"] = cfg

    # setup dababase
    app.on_startup.append(open_db)
    app.on_cleanup.append(close_db)
    return app

def run(cfg):
    # prepare tehe web application
    app = setup_webapp(cfg=cfg)
    
    # run webapp
    aiohttp.web.run_app(host=cfg["local-address"], port=cfg["local-port"], app=app, access_log=logger)



import logging
import sys
import argparse
import pkgutil
import pathlib
import configparser
import os

from pdns_exporter import webapp

logger = logging.getLogger("pdnsexporter")

def setup_cli():
    """setup command-line arguments"""
    options = argparse.ArgumentParser()          
    options.add_argument("-c", help="external config file")   
    options.add_argument('-v', action='store_true', help="verbose mode")

    return options

def setup_logger(cfg):
    loglevel = logging.DEBUG if int(cfg["logs-verbose"]) else logging.INFO
    logfmt = '%(asctime)s %(levelname)s %(message)s'
    
    logger.setLevel(loglevel)
    logger.propagate = False
    
    lh = logging.StreamHandler(stream=sys.stdout )
    lh.setLevel(loglevel)
    lh.setFormatter(logging.Formatter(logfmt))    
    
    logger.addHandler(lh)

def setup_config(args):
    """load default config and update it with arguments or external config"""
    cfg = {}
    parser = configparser.ConfigParser()

    # Set the default configuration file
    content = pkgutil.get_data(__package__, 'settings.conf')
    parser.read_string("[DEFAULT]\n" + content.decode())
    cfg = parser.defaults()

    # Overwrites then with the external file ?    
    if args.c: 
        with open(pathlib.Path(args.c), "r") as fd:
            parser.read_string("[DEFAULT]\n" + fd.read())
            cfg_ext = parser.defaults()
        cfg.update(cfg_ext)

    # Or searches for a file named dnstap.conf in /etc/pdns-exporter/       
    else:
        f = pathlib.Path("/etc/pdns-exporter/settings.conf")
        if f.exists():
            with open(pathlib.Path(args.c), "r") as fd:
                parser.read_string("[DEFAULT]\n" + fd.read())
                cfg_ext = parser.defaults()
            cfg.update(cfg_ext)

    # update verbose mode with command line arguments
    if args.v:
        cfg["logs-verbose"] = args.v

    # finally overwrites config with environment variables
    ENV_VERBOSE = os.getenv('PDNSEXPORT_VERBOSE')
    if ENV_VERBOSE is not None:
        cfg["logs-verbose"] = ENV_VERBOSE

    ENV_LOCAL_ADDRESS = os.getenv('PDNSEXPORT_LOCAL_ADDRESS')
    if ENV_LOCAL_ADDRESS is not None:
        cfg["local-address"] = ENV_LOCAL_ADDRESS

    ENV_LOCAL_PORT = os.getenv('PDNSEXPORT_LOCAL_PORT')
    if ENV_LOCAL_PORT is not None:
        cfg["local-port"] = ENV_LOCAL_PORT

    ENV_API_LOGIN = os.getenv('PDNSEXPORT_API_LOGIN')
    if ENV_API_LOGIN is not None:
        cfg["api-login"] = ENV_API_LOGIN

    ENV_API_PWD = os.getenv('PDNSEXPORT_API_PASSWORD')
    if ENV_API_PWD is not None:
        cfg["api-password"] = ENV_API_PWD

    ENV_DB_HOST = os.getenv('PDNSEXPORT_DB_HOST')
    if ENV_DB_HOST is not None:
        cfg["db-pdns-host"] = ENV_DB_HOST

    ENV_DB_PORT = os.getenv('PDNSEXPORT_DB_PORT')
    if ENV_DB_PORT is not None:
        cfg["db-pdns-port"] = ENV_DB_PORT

    ENV_DB_USER = os.getenv('PDNSEXPORT_DB_USER')
    if ENV_DB_USER is not None:
        cfg["db-pdns-user"] = ENV_DB_USER

    ENV_DB_PWD = os.getenv('PDNSEXPORT_DB_PWD')
    if ENV_DB_PWD is not None:
        cfg["db-pdns-password"] = ENV_DB_PWD

    ENV_DB_NAME = os.getenv('PDNSEXPORT_DB_NAME')
    if ENV_DB_NAME is not None:
        cfg["db-pdns-name"] = ENV_DB_NAME

    return cfg

def start_exporter():
    """start exporter as server"""
    # setup command-line arguments.
    options = setup_cli()
    args = options.parse_args()

    # setup config
    cfg = setup_config(args=args)
    
    # setup logger
    setup_logger(cfg=cfg)

    logger.debug("config OK, starting...")

    # setup webserver
    webapp.run(cfg=cfg)


#!/usr/bin/python

import setuptools
  
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()
    
KEYWORDS = ('powerdns pdns database exporter mysql zone bind')

setuptools.setup(
    name="pdns_exporter",
    version="0.1.0",
    author="Denis MACHARD",
    author_email="d.machard@gmail.com",
    description="PowerDNS records exporter",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/dmachard/python-pdns-exporter",
    packages=['pdns_exporter'],
    include_package_data=True,
    platforms='any',
    keywords=KEYWORDS,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    entry_points={'console_scripts': ['pdns_exporter = pdns_exporter.exporter:start_exporter']},
    install_requires=[
        "SQLAlchemy",
        "aiohttp",
        "aiomysql",
        "requests",
    ]
)